# -*- coding: utf-8 -*-

import os
import sys

# Use local code for local testing, and rely on system-installed module for CI-testing
CI = os.getenv("CI", "false").lower() != "false"
if not CI:
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
    if path not in sys.path:
        sys.path.insert(0, path)

import typeworld.api  # noqa: E402


def Execute(command):
    """\
    Execute system command, return output.
    """

    import os
    import subprocess

    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, shell=True, close_fds=True
    )
    os.waitpid(process.pid, 0)
    response = process.stdout.read().strip()
    process.stdout.close()
    process.wait()
    return response


docstrings = []

docstrings.extend(typeworld.api.RootResponse().docu())
docstrings.extend(typeworld.api.EndpointResponse().docu())
docstrings.extend(typeworld.api.InstallableFontsResponse().docu())
docstrings.extend(typeworld.api.InstallFontsResponse().docu())
docstrings.extend(typeworld.api.UninstallFontsResponse().docu())

docstring = open(os.path.join(os.path.dirname(__file__), "docu.md"), "r").read()


handles = []
for key in [x[0] for x in docstrings]:
    if key not in handles:
        handles.append(key)

classTOC = ""
for handle in handles:
    classTOC += "- [%s](#user-content-class-%s)<br />\n" % (handle, handle.lower())
classTOC += "\n\n"

docstring = docstring.replace("__classTOC__", classTOC)
docstring = docstring.replace("__version__", typeworld.api.VERSION)

# Test code
testCode1Path = os.path.join(os.path.dirname(__file__), "testcode1.py")
docstring = docstring.replace("__testcode1__", open(testCode1Path).read())
docstring = docstring.replace(
    "__testcode1result__", Execute("python3 " + testCode1Path).decode()
)

testCode2Path = os.path.join(os.path.dirname(__file__), "testcode2.py")
docstring = docstring.replace("__testcode2__", open(testCode2Path).read())
docstring = docstring.replace(
    "__testcode2result__", Execute("python3 " + testCode2Path).decode()
)

# Breaking changes
changesList = [f"* `{x}`" for x in typeworld.api.BREAKINGVERSIONS]
docstring = docstring.replace("__breakingAppVersions__", "\n".join(changesList))


for handle in handles:
    for className, string in docstrings:
        if handle == className:
            docstring += string
            docstring += "\n\n"
            break
docstring = docstring.strip() + "\n"

if "TRAVIS" not in os.environ:
    f = open(os.path.join(os.path.dirname(__file__), "README.md"), "w")
    f.write(docstring)
    f.close()



# Import module
import typeworld.api

# Root of Response
root = typeworld.api.RootResponse()

# Response for 'availableFonts' command
installableFonts = typeworld.api.InstallableFontsResponse()
installableFonts.response = "success"

###

# Add designer to root of response
designer = typeworld.api.Designer()
designer.keyword = "max"
designer.name.en = "Max Mustermann"
installableFonts.designers.append(designer)

# Add foundry to root of response
foundry = typeworld.api.Foundry()
foundry.name.en = "Awesome Fonts"
foundry.website = "https://awesomefonts.com"
foundry.uniqueID = "awesomefontsfoundry"
installableFonts.foundries.append(foundry)

# Add license to foundry
license = typeworld.api.LicenseDefinition()
license.keyword = "awesomeFontsEULA"
license.name.en = "Awesome Fonts Desktop EULA"
license.URL = "https://awesomefonts.com/EULA/"
foundry.licenses.append(license)

# Add font family to foundry
family = typeworld.api.Family()
family.name.en = "Awesome Sans"
family.designerKeywords.append("max")
family.uniqueID = "awesomefontsfoundry-awesomesans"
foundry.families.append(family)

# Add version to font family
version = typeworld.api.Version()
version.number = 0.1
family.versions.append(version)

# Add font to family
font = typeworld.api.Font()
font.name.en = "Regular"
font.postScriptName = "AwesomeSans-Regular"
font.licenseKeyword = "awesomeFontsEULA"
font.purpose = "desktop"
font.format = "otf"
font.uniqueID = "awesomefontsfoundry-awesomesans-regular"
family.fonts.append(font)

# Font's license usage
licenseUsage = typeworld.api.LicenseUsage()
licenseUsage.keyword = "awesomeFontsEULA"
font.usedLicenses.append(licenseUsage)

###

# Attach EndpointResponse to RootResponse
root.installableFonts = installableFonts

# Print API response as JSON
print(root.dumpJSON())



import typeworld.client.protocols
import typeworld.api
import requests
import requests.exceptions
from typeworld.api import VERSION


def readJSONResponse(url, responses, acceptableMimeTypes, data={}):
    d = {}
    d["errors"] = []
    d["warnings"] = []
    d["information"] = []

    root = typeworld.api.RootResponse()

    if "source" not in data:
        data["source"] = "typeworldApp"

    # gather commands
    commands = [response._command["keyword"] for response in responses]
    data["commands"] = ",".join(commands)

    try:
        response = requests.post(url, data, timeout=30)
    except requests.exceptions.ConnectionError:
        d["errors"].append(f"Connection refused: {url}")
        return root, d
    except requests.exceptions.HTTPError:
        d["errors"].append(f"HTTP Error: {url}")
        return root, d
    except requests.exceptions.Timeout:
        d["errors"].append(f"Connection timed out: {url}")
        return root, d
    except requests.exceptions.TooManyRedirects:
        d["errors"].append(f"Too many redirects: {url}")
        return root, d

    if response.status_code != 200:
        d["errors"].append(f"HTTP Error {response.status_code}")

    if response.status_code == 200:

        incomingMIMEType = response.headers["content-type"].split(";")[0]
        if incomingMIMEType not in acceptableMimeTypes:
            d["errors"].append(
                "Resource headers returned wrong MIME type: '%s'. Expected is '%s'."
                % (response.headers["content-type"], acceptableMimeTypes)
            )

        # Catching ValueErrors
        try:
            root.loadJSON(response.text)
            information, warnings, errors = root.validate()

            if information:
                d["information"].extend(information)
            if warnings:
                d["warnings"].extend(warnings)
            if errors:
                d["errors"].extend(errors)
        except ValueError as e:
            d["errors"].append(str(e))

    return root, d


class TypeWorldProtocol(typeworld.client.protocols.TypeWorldProtocolBase):
    def initialize(self):
        self.versions = []
        self._endpointCommand = None
        self._installableFontsCommand = None
        self._installFontsCommand = None

    def loadFromDB(self):
        """Overwrite this"""

        if self.get("endpoint"):
            api = typeworld.api.EndpointResponse()
            api.parent = self
            api.loadJSON(self.get("endpoint"))
            self._endpointCommand = api

        if self.get("installableFonts"):
            api = typeworld.api.InstallableFontsResponse()
            api.parent = self
            api.loadJSON(self.get("installableFonts"))
            self._installableFontsCommand = api

        if self.get("installFonts"):
            api = typeworld.api.InstallFontsResponse()
            api.parent = self
            api.loadJSON(self.get("installFonts"))
            self._installFontsCommand = api

    def latestVersion(self):
        return self._installableFontsCommand

    def returnRootCommand(self, testScenario):

        if not self._rootCommand:

            # Read response
            data = {
                "subscriptionID": self.url.subscriptionID,
                "appVersion": VERSION,
            }
            if testScenario:
                data["testScenario"] = testScenario
            root, responses = readJSONResponse(
                self.connectURL(),
                [typeworld.api.EndpointResponse()],
                typeworld.api.INSTALLABLEFONTSCOMMAND["acceptableMimeTypes"],
                data=data,
            )

            # Errors
            if responses["errors"]:
                return False, responses["errors"][0]

            self._endpointCommand = root.endpoint
            self._rootCommand = root

        # Success
        return True, self._rootCommand

    def returnEndpointCommand(self, testScenario):

        if not self._endpointCommand:

            # Read response
            data = {
                "subscriptionID": self.url.subscriptionID,
                "appVersion": VERSION,
            }
            if testScenario:
                data["testScenario"] = testScenario
            root, responses = readJSONResponse(
                self.connectURL(),
                [typeworld.api.EndpointResponse()],
                typeworld.api.INSTALLABLEFONTSCOMMAND["acceptableMimeTypes"],
                data=data,
            )

            # Errors
            if responses["errors"]:
                return False, responses["errors"][0]

            self._endpointCommand = root.endpoint
            self._rootCommand = root

        # Success
        return True, self._endpointCommand

    def returnInstallableFontsCommand(self):
        return True, self.latestVersion()

    def protocolName(self):
        return "Type.World JSON Protocol"

    def update(self):

        data = {
            "subscriptionID": self.url.subscriptionID,
            "anonymousAppID": self.client.anonymousAppID(),
            "anonymousTypeWorldUserID": self.client.user(),
            "appVersion": VERSION,
            "mothership": self.client.mothership,
        }
        if self.client.testScenario:
            data["testScenario"] = self.client.testScenario
        secretKey = self.getSecretKey()
        if secretKey:
            data["secretKey"] = secretKey
        if self.client.testing:
            data["testing"] = "true"

        root, responses = readJSONResponse(
            self.connectURL(),
            [
                typeworld.api.EndpointResponse(),
                typeworld.api.InstallableFontsResponse(),
            ],
            typeworld.api.INSTALLABLEFONTSCOMMAND["acceptableMimeTypes"],
            data=data,
        )

        if responses["errors"]:

            if self.url.unsecretURL() in self.subscription.parent._updatingSubscriptions:
                self.subscription.parent._updatingSubscriptions.remove(self.url.unsecretURL())
            self.subscription._updatingProblem = "\n".join(responses["errors"])
            return False, self.subscription._updatingProblem, False

        if root.installableFonts.response == "error":
            if self.url.unsecretURL() in self.subscription.parent._updatingSubscriptions:
                self.subscription.parent._updatingSubscriptions.remove(self.url.unsecretURL())
            self.subscription._updatingProblem = root.installableFonts.errorMessage
            return False, self.subscription._updatingProblem, False

        if root.installableFonts.response in (
            "temporarilyUnavailable",
            "insufficientPermission",
            "loginRequired",
        ):
            if self.url.unsecretURL() in self.subscription.parent._updatingSubscriptions:
                self.subscription.parent._updatingSubscriptions.remove(self.url.unsecretURL())
            self.subscription._updatingProblem = [
                f"#(response.{root.installableFonts.response})",
                f"#(response.{root.installableFonts.response}.headline)",
            ]
            return (
                False,
                [
                    f"#(response.{root.installableFonts.response})",
                    f"#(response.{root.installableFonts.response}.headline)",
                ],
                False,
            )

        # Security check: Does url begin with canonicalURL?
        if not self.url.HTTPURL().startswith(root.endpoint.canonicalURL):
            return False, "'url' must begin with 'canonicalURL'", None

        # # Detect installed fonts now not available in subscription anymore and delete
        # them
        # hasFonts = False
        # removeFonts = []
        # if api.response == NOFONTSAVAILABLE:
        # 	for foundry in self._installableFontsCommand.foundries:
        # 		for family in foundry.families:
        # 			for font in family.fonts:
        # 				hasFonts = True
        # 				removeFonts.append(font.uniqueID)
        # 	if removeFonts:
        # 		success, message = self.subscription.removeFonts(removeFonts)
        # 		if not success:
        # 			return False, 'Couldn’t uninstall previously installed fonts: %s' %
        # message, True

        # Previously available fonts
        oldIDs = []
        for foundry in self._installableFontsCommand.foundries:
            for family in foundry.families:
                for font in family.fonts:
                    oldIDs.append(font.uniqueID)

        # Newly available fonts
        newIDs = []
        for foundry in root.installableFonts.foundries:
            for family in foundry.families:
                for font in family.fonts:
                    newIDs.append(font.uniqueID)

        # These fonts are no longer available, so delete them.
        # Shrink the list of deletable fonts to the ones actually installed.
        deletedFonts = list(set(oldIDs) - set(newIDs))
        deleteTheseFonts = []
        if deletedFonts:
            for fontID in deletedFonts:
                for foundry in self._installableFontsCommand.foundries:
                    for family in foundry.families:
                        for font in family.fonts:
                            if font.uniqueID == fontID:
                                if self.subscription.installedFontVersion(font.uniqueID):
                                    deleteTheseFonts.append(fontID)

            success, message = self.subscription.removeFonts(deleteTheseFonts, updateSubscription=False)
            if not success:
                return (
                    False,
                    "Couldn’t uninstall previously installed fonts: %s" % message,
                    True,
                )

        # Compare
        changes = self._installableFontsCommand.getContentChanges(root.installableFonts)

        # Success
        self._installableFontsCommand = root.installableFonts

        # EndpointResponse
        if root.endpoint:
            self._endpointCommand = root.endpoint

        # InstallFontsResponse
        if root.installFonts:
            self._installFontsCommand = root.installFonts
        else:
            self._installFontsCommand = None

        self.save()

        return True, None, changes

    def setInstallableFontsCommand(self, command):
        self._installableFontsCommand = command

    # 		self.client.delegate.subscriptionWasUpdated(self.subscription.parent,
    # self.subscription)

    def removeFonts(self, fonts, updateSubscription=False):

        data = {
            "fonts": ",".join(fonts),
            "anonymousAppID": self.client.anonymousAppID(),
            "anonymousTypeWorldUserID": self.client.user(),
            "subscriptionID": self.url.subscriptionID,
            "secretKey": self.getSecretKey(),
            "APIKey": self.client.secretTypeWorldAPIKey,
            "appVersion": VERSION,
            "mothership": self.client.mothership,
        }
        if self.client.testScenario:
            data["testScenario"] = self.client.testScenario
        if self.client.testing:
            data["testing"] = "true"

        commands = [typeworld.api.UninstallFontsResponse()]
        if updateSubscription:
            commands.append(typeworld.api.InstallableFontsResponse())

        root, messages = readJSONResponse(
            self.connectURL(),
            commands,
            typeworld.api.UNINSTALLFONTSCOMMAND["acceptableMimeTypes"],
            data=data,
        )
        api = root.uninstallFonts

        if messages["errors"]:
            return False, "\n\n".join(messages["errors"])

        if api.response == "error":
            return False, api.errorMessage

        if api.response != "success":
            return (
                False,
                [f"#(response.{api.response})", f"#(response.{api.response}.headline)"],
            )

        if updateSubscription and root.installableFonts and root.installableFonts.response == "success":
            self.setInstallableFontsCommand(root.installableFonts)

        return True, api

    def installFonts(self, fonts, updateSubscription=False):

        # Build URL

        if self._installFontsCommand:
            return True, self._installFontsCommand

        else:

            data = {
                "fonts": ",".join([f"{x[0]}/{x[1]}" for x in fonts]),
                "anonymousAppID": self.client.anonymousAppID(),
                "anonymousTypeWorldUserID": self.client.user(),
                "subscriptionID": self.url.subscriptionID,
                "secretKey": self.getSecretKey(),
                "APIKey": self.client.secretTypeWorldAPIKey,
                "appVersion": VERSION,
                "mothership": self.client.mothership,
            }
            if self.client.testScenario:
                data["testScenario"] = self.client.testScenario
            if self.client.testing:
                data["testing"] = "true"

            if self.subscription.get("revealIdentity") and self.client.userName():
                data["userName"] = self.client.userName()
            if self.subscription.get("revealIdentity") and self.client.userEmail():
                data["userEmail"] = self.client.userEmail()

            commands = [typeworld.api.InstallFontsResponse()]
            if updateSubscription:
                commands.append(typeworld.api.InstallableFontsResponse())

            root, messages = readJSONResponse(
                self.connectURL(),
                commands,
                typeworld.api.INSTALLFONTSCOMMAND["acceptableMimeTypes"],
                data=data,
            )
            api = root.installFonts

            if messages["errors"]:
                return False, "\n\n".join(messages["errors"])

            if api.response == "error":
                return False, api.errorMessage

            if api.response != "success":
                return (
                    False,
                    [
                        f"#(response.{api.response})",
                        f"#(response.{api.response}.headline)",
                    ],
                )

            if updateSubscription and root.installableFonts and root.installableFonts.response == "success":
                self.setInstallableFontsCommand(root.installableFonts)

            return True, api

    def aboutToAddSubscription(
        self,
        anonymousAppID,
        anonymousTypeWorldUserID,
        accessToken,
        testScenario,
    ):
        """Overwrite this.
        Put here an initial health check of the subscription. Check if URLs point to the
        right place etc.
        Return False, 'message' in case of errors."""

        # Read response
        data = {
            "subscriptionID": self.url.subscriptionID,
            "secretKey": self.url.secretKey,
            "anonymousAppID": anonymousAppID,
            "anonymousTypeWorldUserID": anonymousTypeWorldUserID,
            "accessToken": accessToken,
            "APIKey": self.client.secretTypeWorldAPIKey,
            "appVersion": VERSION,
            "mothership": self.client.mothership,
        }
        if testScenario:
            data["testScenario"] = testScenario
        if self.client.testing:
            data["testing"] = "true"

        root, responses = readJSONResponse(
            self.connectURL(),
            [
                typeworld.api.EndpointResponse(),
                typeworld.api.InstallableFontsResponse(),
            ],
            typeworld.api.INSTALLABLEFONTSCOMMAND["acceptableMimeTypes"],
            data=data,
        )

        # Errors
        if responses["errors"]:
            return False, responses["errors"][0]

        if (
            "installableFonts" not in root.endpoint.supportedCommands
            or "installFonts" not in root.endpoint.supportedCommands
        ):
            return (
                False,
                "API endpoint %s does not support the 'installableFonts' and 'installFonts' commands."
                % root.endpoint.canonicalURL,
            )

        if root.installableFonts.response == "error":
            if root.installableFonts.errorMessage:
                return False, root.installableFonts.errorMessage
            else:
                return False, "api.response is error, but no error message was given"

        # Predefined response messages
        if root.installableFonts.response != "error" and root.installableFonts.response != "success":
            return (
                False,
                [
                    "#(response.%s)" % root.installableFonts.response,
                    "#(response.%s.headline)" % root.installableFonts.response,
                ],
            )

        # Security check: Does url begin with canonicalURL?
        if not self.url.HTTPURL().startswith(root.endpoint.canonicalURL):
            return False, "'url' must begin with 'canonicalURL'"

        # Success
        self._rootCommand = root
        self._endpointCommand = root.endpoint
        self._installableFontsCommand = root.installableFonts

        # InstallFontsResponse
        if root.installFonts:
            self._installFontsCommand = root.installFonts
        else:
            self._installFontsCommand = None

        if self.url.secretKey:
            self.setSecretKey(self.url.secretKey)

        return True, None

    def save(self):

        assert self._endpointCommand
        self.set("endpoint", self._endpointCommand.dumpJSON(validate=False))

        assert self._installableFontsCommand
        self.set("installableFonts", self._installableFontsCommand.dumpJSON(validate=False))

        if self._installFontsCommand:
            self.set("installFonts", self._installFontsCommand.dumpJSON(validate=False))
        else:
            self.set("installFonts", "")



import platform
import os
import sys
import traceback
import time
import subprocess


def ReadFromFile(path):
    """\
    Return content of file
    """
    import os
    import codecs

    if os.path.exists(path):
        f = codecs.open(path, encoding="utf-8", mode="r")
        text = f.read()  # .decode('utf8')
        f.close()
        return text


def WriteToFile(path, string):
    """\
    Write content to file
    """
    f = open(path, "wb")
    f.write(string.encode())
    f.close()
    return True


def Execute(command):
    """\
    Execute system command, return output.
    """

    import os
    import subprocess

    process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True, close_fds=True)
    os.waitpid(process.pid, 0)
    response = process.stdout.read().strip()
    process.stdout.close()
    process.wait()
    return response


def Garbage(length, uppercase=True, lowercase=True, numbers=True, punctuation=False):
    """\
    Return string containing garbage.
    """

    import random

    uppercaseparts = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    lowercaseparts = "abcdefghijklmnopqrstuvwxyz"
    numberparts = "0123456789"
    punctuationparts = ".-_"

    pool = ""
    if uppercase:
        pool += uppercaseparts
    if lowercase:
        pool += lowercaseparts
    if numbers:
        pool += numberparts
    if punctuation:
        pool += punctuationparts

    garbage = ""

    while len(garbage) < length:
        garbage += random.choice(pool)

    return garbage


def get_registry_value(key, subkey, value):
    import winreg

    key = getattr(winreg, key)
    handle = winreg.OpenKey(key, subkey)
    (value, type) = winreg.QueryValueEx(handle, value)
    return value


def install_font(path, data):
    try:
        f = open(path, "wb")
        f.write(data)
        f.close()

        if WIN:
            register_font_in_winreg(path)

        return True, None

    except Exception:
        return False, traceback.format_exc()


def uninstall_font(path):
    if WIN:
        unregister_font_in_winreg(path)

    i = 0
    while os.path.exists(path) and i < 100:
        try:
            os.remove(path)
        except PermissionError:
            if WIN:
                print("delete using secondary procedure")
                subprocess.call(f"del {path}", shell=True)
            time.sleep(0.1)
        i += 1

    # Font still not deleted
    if os.path.exists(path):

        # Put registry entry bacl where it was
        if WIN:
            register_font_in_winreg(path)

        return (
            False,
            "Font couldn’t be removed.",
        )

    return True, None


def register_font_in_winreg(path):
    from fontTools.ttLib import TTFont

    f = TTFont(path)
    name = (f["name"].getName(4, 3, 1, 1033) or f["name"].getName(4, 1, 0, 0)).toUnicode()
    string = f"{name} ({os.path.splitext(path)[-1].split('.')[-1]})"
    f.close()
    import winreg as wreg

    key = wreg.OpenKey(
        wreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Fonts", 0, wreg.KEY_SET_VALUE
    )
    wreg.SetValueEx(key, string, 0, wreg.REG_SZ, os.path.basename(path))


def unregister_font_in_winreg(path):
    from fontTools.ttLib import TTFont

    f = TTFont(path)
    name = (f["name"].getName(4, 3, 1, 1033) or f["name"].getName(4, 1, 0, 0)).toUnicode()
    string = f"{name} ({os.path.splitext(path)[-1].split('.')[-1]})"
    f.close()
    import winreg as wreg

    key = wreg.OpenKey(
        wreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Fonts", 0, wreg.KEY_ALL_ACCESS
    )
    try:
        wreg.DeleteValue(key, string)
    except FileNotFoundError:
        pass
    key.Close()


# https://www.tenforums.com/tutorials/54452-rebuild-font-cache-windows-10-a.html
def rebuild_font_cache():
    os.system('sc stop "FontCache"')
    os.system('sc config "FontCache" start=disabled')
    os.system('del /A /F /Q "%WinDir%\\ServiceProfiles\\LocalService\\AppData\\Local\\FontCache\\*FontCache*"')
    os.system('del /A /F /Q "%WinDir%\\System32\\FNTCACHE.DAT"')
    os.system('sc config "FontCache" start=auto')
    os.system('sc start "FontCache"')


def MachineName():

    machineModelIdentifier = None
    humanReadableName = None
    specsDescription = None

    if platform.system() == "Windows":

        specsDescription = get_registry_value(
            "HKEY_LOCAL_MACHINE",
            "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
            "ProcessorNameString",
        )
        humanReadableName = get_registry_value(
            "HKEY_LOCAL_MACHINE",
            "HARDWARE\\DESCRIPTION\\System\\BIOS",
            "SystemProductName",
        )

    if platform.system() == "Linux":

        cpu = ""
        itemsUsed = []
        procinfo = Execute("cat /proc/cpuinfo").decode()

        for line in procinfo.split("\n"):
            if ":" in line:
                k, v = line.split(":")[:2]
                if k.strip() == "model name" and k not in itemsUsed:
                    cpu += v.strip()
                    itemsUsed.append(k)

        if os.path.exists("/sys/devices/virtual/dmi/id/sys_vendor") and os.path.exists(
            "/sys/devices/virtual/dmi/id/product_name"
        ):
            humanReadableName = "%s %s" % (
                Execute("cat /sys/devices/virtual/dmi/id/sys_vendor").decode(),
                Execute("cat /sys/devices/virtual/dmi/id/product_name").decode(),
            )
        else:
            humanReadableName = "Google App Engine"  # nocoverage
        specsDescription = cpu

    elif platform.system() == "Darwin":

        name = None

        # Approach 1
        import plistlib

        data = plistlib.loads(Execute("system_profiler -xml SPHardwareDataType"))

        # Approach 2
        if not name:
            name = data[0]["_items"][0]["machine_name"]

        machineModelIdentifier = data[0]["_items"][0]["machine_model"]
        humanReadableName = "%s" % name
        # if not name.startswith('Apple'):
        # 	name = 'Apple ' + name
        specsDescription = []
        # if 'cpu_type' in data[0]['_items'][0]:
        # 	specsDescription.append(data[0]['_items'][0]['cpu_type'])
        if "current_processor_speed" in data[0]["_items"][0]:
            specsDescription.append(data[0]["_items"][0]["current_processor_speed"])
        specsDescription.append("with")
        if "physical_memory" in data[0]["_items"][0]:
            specsDescription.append(data[0]["_items"][0]["physical_memory"])

        specsDescription = " ".join(specsDescription)

    return machineModelIdentifier, humanReadableName, specsDescription


def OSName():

    import platform

    if platform.system() == "Darwin":
        return "macOS %s" % platform.mac_ver()[0]

    elif platform.system() == "Windows":

        return (
            get_registry_value(
                "HKEY_LOCAL_MACHINE",
                "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",
                "ProductName",
            )
            + ", "
            + str(platform.version())
        )

    elif platform.system() == "Linux":
        return " ".join(platform.platform())


def addAttributeToURL(url, attributes):

    from urllib.parse import urlparse

    o = urlparse(url)

    for attribute in attributes.split("&"):

        key, value = attribute.split("=")

        replaced = False
        queryParts = o.query.split("&")
        if queryParts:
            for i, query in enumerate(queryParts):
                if "=" in query and query.startswith(key + "="):
                    queryParts[i] = attribute
                    replaced = True
                    break
        if not replaced:
            if queryParts[0]:
                queryParts.append(attribute)
            else:
                queryParts[0] = attribute
        o = o._replace(query="&".join(queryParts))

    return o.geturl()


WIN = platform.system() == "Windows"
MAC = platform.system() == "Darwin"
LINUX = platform.system() == "Linux"


# from https://gist.github.com/dgelessus/018681c297dfae22a06a3bc315d9e6e3:

"""Allows logging Python :class:`str`s to the system log using Foundation's ``NSLog``
function. This module is meant to help with debugging the loading process of Rubicon on
systems like iOS, where sometimes only NSLog output (but not stdout/stderr) is available
to the developer."""

if MAC:
    import ctypes

    __all__ = [
        "nslog",
    ]

    # Name of the UTF-16 encoding with the system byte order.
    if sys.byteorder == "little":
        UTF16_NATIVE = "utf-16-le"
    elif sys.byteorder == "big":  # nocoverage
        UTF16_NATIVE = "utf-16-be"  # nocoverage
    else:
        raise AssertionError("Unknown byte order: " + sys.byteorder)  # nocoverage

    # Note: Many of the following definitions are duplicated in other rubicon.objc
    # submodules.
    # However because this module is used to debug the early startup process of Rubicon,
    # we can't use any of the other
    # definitions.

    class CFTypeRef(ctypes.c_void_p):
        pass

    CFIndex = ctypes.c_long
    UniChar = ctypes.c_uint16

    CoreFoundation = ctypes.CDLL("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")
    Foundation = ctypes.CDLL("/System/Library/Frameworks/Foundation.framework/Foundation")

    # void CFRelease(CFTypeRef arg)
    CoreFoundation.CFRelease.restype = None
    CoreFoundation.CFRelease.argtypes = [CFTypeRef]

    # CFStringRef CFStringCreateWithCharacters(CFAllocatorRef alloc, const UniChar
    # *chars, CFIndex numChars)
    CoreFoundation.CFStringCreateWithCharacters.restype = CFTypeRef
    CoreFoundation.CFStringCreateWithCharacters.argtypes = [
        CFTypeRef,
        ctypes.POINTER(UniChar),
        CFIndex,
    ]

    # void NSLog(NSString *format, ...)
    Foundation.NSLog.restype = None
    Foundation.NSLog.argtypes = [CFTypeRef]

    def _cfstr(s):
        """Create a ``CFString`` from the given Python :class:`str`."""

        encoded = s.encode(UTF16_NATIVE)
        assert len(encoded) % 2 == 0
        arr = (UniChar * (len(encoded) // 2)).from_buffer_copy(encoded)
        cfstring = CoreFoundation.CFStringCreateWithCharacters(None, arr, len(arr))
        assert cfstring is not None
        return cfstring

    # Format string for a single object.
    FORMAT = _cfstr("%@")

    def nslog(s):
        """Log the given Python :class:`str` to the system log."""

        cfstring = _cfstr(s)
        Foundation.NSLog(FORMAT, cfstring)
        CoreFoundation.CFRelease(cfstring)

    # class NSLogWriter(io.TextIOBase):
    # 	"""An output-only text stream that writes to the system log."""

    # 	def write(self, s):
    # 		nslog(s[:-1] if s.endswith('\n') else s)
    # 		return len(s)

    # 	@property
    # 	def encoding(self):
    # 		return UTF16_NATIVE
